''=============================================================================
'' @file     sht2x_demo.spin    demo of basic funtionality
'' @target   Propeller with Sensirion SHT2x
'' @author   Thomas Tracy Allen, EME Systems
'' Copyright (c) 2013,2015 EME Systems LLC
'' See end of file for terms of use.
'' version 1.0
''
''=============================================================================

CON
  _clkmode = xtal1 + pll8x                '
  _xinfreq = 5_000_000



  SCL = 9
  SDA = 8                       '

OBJ
  sht : "sht21_object"
  uarts : "fullduplexserial4port"

           
VAR

PUB demo | degC, RH
  Startuarts
  uarts.tx(0,13)
  uarts.str(0,string(13,"--top of SHT21 demo"))

  if (result:=sht.Init(SCL, SDA)) <> 6   ' selects i2c pins and checks for pullups and ACK after query.
    uarts.Str(0,string(13,"--can't find SHT2x sensor, status: ",32))
    uarts.bin(0,result,3)
    repeat
  sht.SetResolution(sht#LOW)   ' default resolution is sht#HIGH.   sht#LOW is faster and adequate resolution.

  repeat
    uarts.char(0,13)

  '--the following two methods use the clock hold mechanism of the sht21
    degC := sht.readTemperature
    uarts.decPW(0,degC,2,7)

    RH := sht.readHumidity
    uarts.decPW(0,RH,1,7)

  '--alternative method below use polling or delay to wait for data ready
    sht.TriggerRead(sht#TC)
    repeat
    until sht.PollReady == 0
    ' Pause(90)   ' alternatively use a delay
    degC := sht.ReadResult
    uarts.decPW(0,degC,2,7)

    sht.TriggerRead(sht#RH)
    Pause(30)
    RH := sht.ReadResult
    uarts.decPW(0,RH,1,7)

    Pause(1000)


PUB Pause(ms)
  waitcnt(clkfreq/1000*ms+cnt)


PUB StartUarts
  uarts.init
  uarts.addport(0,31,30,-1,-1,0,0,9600)
  uarts.start
  Pause(100)



{{
┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │                                                            
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │ 
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
}}
